/*---------------------------------------------------*/
/* File        : nbody.c                             */
/* Description : Computes the correlation function   */
/*               between two sets of values          */
/* Platform    : Sun Ultra Cluster                   */
/*               Virtual Programming Lab             */
/* Application : CPS615 Assignment #4, Fall '98      */
/*                                                   */
/* Author      : Subhash Nair (subhash@cat.syr.edu)  */
/*---------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define N 24

/* The function that is used to compute the correlation */
double f(double a, double b)
{
    return (a*a + b*b);
}


void main(int argc, char* argv[])
{ 
    int n;        /* To store the number of bodies per processor */
    int sendto;   /* Rank of process to which data is to be sent */
    int recvfrom; /* Rank of process from which data is to be received */
    int my_rank;  /* Rank of this process */

    /* Arrays containing the two sets of measurements */
    double global_x[N];
    double global_y[N];

    /* Arrays that will hold the set of values for this process*/
    /*(derived from the global arrays) */
    double local_x[N];
    double local_y[N];

    double local_correlation[N];/* Correlation array for this processor */
    double tempY[N];            /* Temporary array used to receive y's from other processes */

    double sum;         /* To sum up the elements of the correlation array for each processor */
    double correlation; /* To hold the sum of all the correlation functions calculated by each process */
    int p;              /*  The total number of processes in the communicator  */
    int i,j,k;          /* Counters */

    MPI_Status status;
                
    /* Initializes MPI environment using command line arguments */
    MPI_Init(&argc, &argv);

    /* Find the number of processors being used */
    MPI_Comm_size(MPI_COMM_WORLD, &p);

    /* Find the rank of this process */
    MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

    /* Each processor handles N/p values */      
    n = N/p;
        
    /* Initialize the arrays */
    for (i=0; i<N;i++) 
    {
        global_x[i] = (double)(i*i) + N;   /* Arbitrarily generating values*/
        global_y[i] = (double)(i*i) + (N*N);
        local_x[i] = 0.0;
        local_y[i] = 0.0;
        local_correlation[i] = 0.0;
        tempY[i]=0.0;
    }

    j=0;

    /* Decide what values of the global set will be alloted to */
    /* the local set, based on the rank of this process */
    for(i=my_rank*n; i<(my_rank+1)*n; i++)
    {
        local_x[j] = global_x[i];
        local_y[j] = global_y[i];
        j++;
    }

    /* Decide which processor this sends to and receives from  */
    sendto = (my_rank == p-1)?0:(my_rank+1);
    recvfrom = (my_rank == 0)?(p-1):(my_rank-1);

    sum = 0.0; correlation=0.0;   /*  initializing to zero  */
    for (k=0; k<p; k++) 
    {	
        for ( j=0; j<n; j++)
            for ( i=0; i<n; i++)             
                local_correlation[i] += f(local_x[i], local_y[j]);

        MPI_Sendrecv(&local_y, N, MPI_DOUBLE, sendto ,0, &tempY, N, MPI_DOUBLE, recvfrom ,0, MPI_COMM_WORLD, &status); 
                  
        for(i = 0; i < n; i++)	
            local_y[i] = tempY[i];              
    }
     
    /* summing up the elements of the correlation array  by each process  */
    for(i = 0; i < n; i++)
       sum += local_correlation[i];	

    /*A global reduction operation carried out to sum up the correlation functions obtained from all the processors  */
    MPI_Reduce(&sum, &correlation, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

    /*  Process 0 displays the correlation function obtained  */
    if (my_rank == 0)
        printf("Value of correlation function = %.16f\n", correlation);

    MPI_Finalize();
}
